// ----------------------------------------------------------------------------
//
//  Copyright (C) 2003-2018 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <stdlib.h>
#include <string.h>
#include "impdata3.h"


//------------------------------------------------------------------------------------------------------------


const unsigned char Impdata::uuid [16] = {
    0x2b, 0xa2, 0x47, 0xfa, 0x78, 0xbb, 0x4c, 0x5a,
    0x9d, 0xfe, 0x35, 0x68, 0xb1, 0xbd, 0xdc, 0x89 };


Impdata::Impdata (void) :
    _vers (3),
    _type (0),
    _rate_n (0),
    _rate_d (1),
    _n_fram (0),
    _n_sect (0),
    _tref_i (0),
    _tref_n (0),
    _tref_d (1),
    _bits (0),
    _i_fram (0),
    _i_sect (0),
    _mode (0),
    _size (0),
    _data (0),
    _caffile (0),
    _sndfile (0)
{
}


Impdata::~Impdata (void)
{
    close ();
    delete[] _data;
}


int Impdata::set_type (uint32_t type)
{
    if (_mode) return ERR_MODE;
    _type = type;
    deall ();
    return 0;
}


int Impdata::set_n_fram (uint32_t n_fram)
{
    if (_caffile) return ERR_MODE;
    _n_fram = n_fram;
    deall ();
    return 0;
}


int Impdata::set_n_sect (uint32_t n_sect)
{
    if (_caffile) return ERR_MODE;
    _n_sect = n_sect;
    return 0;
}


int Impdata::alloc (void)
{
    uint32_t k;

    k = (_type & TYPE__CHAN) * _n_fram;
    if (k != _size)
    {
        delete[] _data;
	if (k)
	{
	    _size = k;
            _data = new float [k];
            memset (_data, 0, _size * sizeof (float));
	}
        else
	{
            _size = 0;
            _data = 0;  
	}
    }
    return 0;
}


int Impdata::deall (void)
{
    delete[] _data;
    _data = 0;
    _size = 0;
    return 0;
}


int Impdata::clear (void)
{
    if (_data) memset (_data, 0, _size * sizeof (float));
    return 0;
}


//------------------------------------------------------------------------------------------------------------


int Impdata::open_read (const char *name)
{
    int           i, nchk, ichk;
    unsigned char p [40];

    if (_mode) return ERR_MODE;
    deall ();
    _type = 0;
    _n_fram = 0;
    _n_sect = 0;

    _caffile = new Caffile ();
    if (_caffile->open_read (name))
    {
	delete _caffile;
	_caffile = 0;
	return ERR_OPEN;
    }

    nchk = _caffile->nchunks ();
    ichk = 0;
    for (i = 1; i < nchk - 1; i++)
    {
	if (   !memcmp (_caffile->chunk_type (i), "uuid", 4)
	    && _caffile->read_chunk (i, 0, 16, p) == 16
	    && !memcmp (p, uuid, 16))
	{
	    ichk = i;
	    break;
	}
    }

    if (ichk == 0 || _caffile->read_chunk (ichk, 16, 40, p) != 40)
    {
	delete _caffile;
	_caffile = 0;
	return ERR_OPEN;
    }

    _vers   = getbe32 (p + 0);
    _type   = getbe32 (p + 4);
    _rate_n = getbe32 (p + 8);
    _rate_d = getbe32 (p + 12);
    _n_fram = getbe32 (p + 16);
    _n_sect = getbe32 (p + 20);
    _tref_i = getbe32 (p + 24);
    _tref_n = getbe32 (p + 28);
    _tref_d = getbe32 (p + 32);
    _bits   = getbe32 (p + 36);

    if (!_type || !_n_fram || !_n_sect)
    {
	delete _caffile;
	_caffile = 0;
	return ERR_FORM;
    }

    _mode = MODE_READ;
    _i_sect = 0;
    _i_fram = 0;

    return 0;
}


int Impdata::open_write (const char *name)
{
    int           ichk;
    unsigned char p [40];

    if (_mode || !_type || !_n_fram || !_n_sect) return ERR_MODE;

    _caffile = new Caffile ();
    if (_caffile->open_write (name,
	  	              (double)_rate_n / _rate_d,
			      _type & TYPE__CHAN,
			      Caffile::FORM_FLT32 | Caffile::FORM_LE))
    {
	delete _caffile;
	_caffile = 0;
	return ERR_OPEN;
    }

    ichk = _caffile->create_chunk ("uuid", 56);
    _caffile->write_chunk (ichk, 0, 16, uuid);
    putbe32 (p +  0, _vers);
    putbe32 (p +  4, _type);
    putbe32 (p +  8, _rate_n);
    putbe32 (p + 12, _rate_d);
    putbe32 (p + 16, _n_fram);
    putbe32 (p + 20, _n_sect);
    putbe32 (p + 24, _tref_i);
    putbe32 (p + 28, _tref_n);
    putbe32 (p + 32, _tref_d);
    putbe32 (p + 36, _bits);
    _caffile->write_chunk (ichk, 16, 40, p);
    
    _mode = MODE_WRITE;
    _i_sect = 0;
    _i_fram = 0;

    return 0;
}


int Impdata::sf_open_read (const char *name)
{
    SF_INFO   I;

    if (_mode) return ERR_MODE;
    deall ();
    _type = 0;
    _n_fram = 0;
    _n_sect = 0;

    if ((_sndfile = sf_open (name, SFM_READ, &I)) == 0) return ERR_OPEN;

    _type   = TYPE_RAW + I.channels;
    _rate_n = I.samplerate;
    _rate_d = 1;
    _n_sect = 1;
    _n_fram = I.frames;
    _tref_i = 1;
    _tref_n = 0;
    _tref_d = 1;
    _bits   = 0;

    if (sf_command (_sndfile, SFC_WAVEX_GET_AMBISONIC, 0, 0) == SF_AMBISONIC_B_FORMAT)
    {
	switch (I.channels)
	{
	case 3: _type = TYPE_AMB_B3; break;
	case 4: _type = TYPE_AMB_B4; break;
	}
    }

    if (!_n_fram)
    {
	sf_close (_sndfile);
	_sndfile = 0;
	return ERR_FORM;
    }

    _mode = MODE_READ;
    
    return 0;
}


int Impdata::sf_open_write (const char *name)
{
    SF_INFO   I;

    if (_mode || !_type || !_n_fram || !_n_sect) return ERR_MODE;

    I.channels = _type & TYPE__CHAN;
    I.format = (I.channels > 2) ? SF_FORMAT_WAVEX : SF_FORMAT_WAV;
    I.format |= SF_FORMAT_FLOAT;
    I.samplerate = _rate_n / _rate_d;
    I.sections = 1;

    if ((_sndfile = sf_open (name, SFM_WRITE, &I)) == 0) return ERR_OPEN;

    switch (_type)
    {
    case TYPE_AMB_B3:
    case TYPE_AMB_B4:
        sf_command (_sndfile, SFC_WAVEX_SET_AMBISONIC, 0, SF_AMBISONIC_B_FORMAT);
	break;
    }

    _mode = MODE_WRITE;
    _i_sect = 0;
    _i_fram = 0;

    return 0;
}


int Impdata::close (void)
{
    if (_caffile)
    {
	delete _caffile;
        _caffile = 0;
    }
    else if (_sndfile)
    {
	sf_close (_sndfile);
	_sndfile = 0;
    }
    _mode = 0;
    return 0;
}


//------------------------------------------------------------------------------------------------------------


int Impdata::read_sect (uint32_t i_sect)
{
    int r;

    r = locate (i_sect);
    if (r) return r;
    return _data ? read_ext (_n_fram, _data) : ERR_DATA;
}


int Impdata::write_sect (uint32_t i_sect)
{
    int r;

    r = locate (i_sect);
    if (r) return r;
    return _data ? write_ext (_n_fram, _data) : ERR_DATA;
}


int Impdata::locate (uint32_t i_sect)
{
    if (_caffile)
    {
        if (_caffile->seek (_n_fram * i_sect, SEEK_SET) != _n_fram * i_sect) return ERR_SEEK;
        _i_sect = i_sect;
        _i_fram = 0;
        return 0;
    }
    else if (_sndfile)
    {
	if (sf_seek (_sndfile, _n_fram * i_sect, SEEK_SET) != _n_fram * i_sect) return ERR_SEEK;
        _i_sect = i_sect;
        _i_fram = 0;
	return 0;
    }
    else return ERR_MODE;
}


int Impdata::read_ext (uint32_t k, float *p)
{
    if (_mode != MODE_READ) return ERR_MODE;
    if (k > _n_fram - _i_fram) k = _n_fram - _i_fram;
    if (_caffile)
    {
        k = _caffile->read (k, p);
    }
    else if (_sndfile)
    {
	k = sf_readf_float (_sndfile, p, k);
    }
    else return ERR_MODE;
    _i_fram += k;
    return k;
}


int Impdata::write_ext (uint32_t k, float *p)
{
    if (_mode != MODE_WRITE) return ERR_MODE;
    if (k > _n_fram - _i_fram) k = _n_fram - _i_fram;
    if (_caffile)
    {
        k = _caffile->write (k, p);
    }
    else if (_sndfile)
    {
	k = sf_writef_float (_sndfile, p, k);
    }
    else return ERR_MODE;
    _i_fram += k;
    return k;
}


//------------------------------------------------------------------------------------------------------------


void Impdata::checkext (char *name)
{
    char  *p;

    p = strrchr (name, '.');
    if (p && ! strcmp (p, ".ald")) return;
    strcat (name, ".ald");
}


char Impdata::_suffix [12];
const char *Impdata::_suffix2 [] = { "F", "I" };
const char *Impdata::_suffix3 [] = { "L", "R" };
const char *Impdata::_suffix4 [] = { "M", "S" };
const char *Impdata::_suffix5 [] = { "LF", "RF", "LB", "RB" };
const char *Impdata::_suffix7 [] = { "W", "X", "Y", "Z" }; 


const char *Impdata::channame (uint32_t type, uint32_t chan)
{
    const char **p;

    if (chan >= (type & TYPE__CHAN)) return 0;

    switch (type & TYPE__TYPE)
    {
    case 0x00020000: p = _suffix2; break;
    case 0x00030000: p = _suffix3; break;
    case 0x00040000: p = _suffix4; break;
    case 0x00050000: p = _suffix5; break;
    case 0x00060000:
    case 0x00070000: p = _suffix7; break;
    default: p = 0;	
    }

    if (p) return p [chan];
    sprintf (_suffix, "%d", chan + 1);
    return _suffix;
}


//------------------------------------------------------------------------------------------------------------
